@@ -1,4 +1,24 @@
-/* $Id: opt_proc.c,v 3.5 1997/11/02 00:23:58 jt Exp $ */
+/* OPT v3.7: options parsing tool */
+/*
+ * Copyright (C) 1998, James Theiler (jt@lanl.gov)
+ * 
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License as published by the Free Software Foundation; either
+ * version 2 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the
+ * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+ * Boston, MA  02111-1307, USA.
+ * 
+ */
+/* $Id: opt_proc.c,v 3.7 1998/07/25 19:45:11 jt Exp $ */
 /* opt_proc.c */
 /*
  *	process options in the OPT package
@@ -27,14 +47,16 @@
  */
 
 char	*opt_program_name = NULL;
-flag	fileflag = False;
+flag	opt_fileflag = False;
 
-char	*optfile_name = NULL; 
+char	*optfile_name = NULL;
+char    *opt_defaultfile = NULL;
 char    *opt_titlestring = NULL;
 char    *opt_usagestring = NULL;
 char	*opt_envstring=NULL;
 char	*opt_defstring=NULL;
 
+OPT_PFI     opt_additional_usage_fcn=NULL;
 OPT_PFI	    opt_quit_fcn=NULL;
 OPT_PFI	    opt_run_fcn=NULL;
 OPT_PFI_ARG opt_main_fcn=NULL;
@@ -62,8 +84,10 @@ optTitle(char *s)
 char *
 optgetTitle(void)
 {
-    if (opt_titlestring) return opt_titlestring;
-    else       return opt_program_name;
+    if (!ISEMPTYSTRING(opt_titlestring))
+        return opt_titlestring;
+    else
+        return opt_program_name;
 }
 void
 optUsage(char *s)
@@ -81,6 +105,11 @@ optDefaultString(char *s)
     opt_defstring = strdup(s);
 }
 void
+optDefaultFile(char *s)
+{
+    opt_defaultfile = strdup(s);
+}
+void
 optQuit(OPT_PFI fcn)
 {
     opt_quit_fcn = fcn;
@@ -95,17 +124,21 @@ optMain(OPT_PFI_ARG fcn)
 {
     opt_main_fcn = fcn;
 }
+void
+optAdditionalUsage(OPT_PFI fcn)
+{
+    opt_additional_usage_fcn = fcn;
+}
+
 
 
 /* ------------------------------------------------------------------ */
 
-	/* ------- */
-	/* getopts */
-	/* ------- */
-
- /*
- * this is the routine that is called from main(argc,argv).
- * returns value of argc.
+/* getopts:
+ * This routine is maintained only for backward compatibility.  It
+ * takes argc, argv and returns the new value of argc; Now, it is
+ * recommended that you use opt(), which takes &argc,&argv as arguments,
+ * and then returns new argc,argv through the pointers on the arg list.
  */
 int
 getopts(int argc, char **argv)
@@ -140,9 +173,10 @@ getopts(int argc, char **argv)
 
     return(argc);
 }
- /*
- * this is an alternative routine that can be called from main(argc,argv).
- * it returns nothing, but argc/argv are returned via the arglist
+/*
+ * opt: This is the routine that does all the work.  It is generally called
+ * from main(), after all calls to optreg() have been made.
+ * It returns nothing, but new argc/argv are returned via the arglist
  */
 void
 opt(int *argcptr, char ***argvptr)
@@ -169,7 +203,12 @@ opt(int *argcptr, char ***argvptr)
     /* First process default string */
     opt_lineprocess(opt_defstring);
 
-    /* Second, check environment variable */
+    /* Second, check the default file (eg ~/.programrc) */
+    if (!ISEMPTYSTRING(opt_defaultfile)) {
+        opt_fromfname(opt_defaultfile);
+    }
+    
+    /* Third, check environment variable */
     if(opt_envstring != NULL)
     {
 	char	*s;
@@ -350,21 +389,17 @@ opt_process(int argc, char **argv)
             break;
 
         case OPTFROMFILE:
-            if(1) {
-                flag tmp_fileflag;
-                tmp_fileflag = fileflag;
-                fileflag = True;
-                opt_fromfname(argnext(ag));
-                fileflag = tmp_fileflag;
+            if (! opt_fromfname(argnext(ag))) {
+                opt_warn_1("Cannot open options file [%s]\n",optfile_name);
             }
             break;
-
+            
         case OPTTOFILE:
             opt_tofname(argnext(ag));
             break;
 
         case HELPCH:
-            if( !fileflag)
+            if( !opt_fileflag)
                 opt_help(argnext(ag));
             break;
 
@@ -679,23 +714,26 @@ opt_fromfname(char *fname)
 {
     FILE	*fp;
     char	line[MAXOPLINE+2];	/* leave room for termination */
-
+    flag        tmp_fileflag;
+    
     if (ISEMPTYSTRING(fname)) {
         opt_warn_1("%cfile.opt fails to specify file.opt\n",OPTFROMFILE);
         return 0;
     }
-    if (*fname != OPTFROMFILE) {      /* if not default '@@' */
-        optfile_name = strdup(fname); /* then reset default to file.opt */
+    if (*fname == OPTFROMFILE) {      /* if '@@' */
+        fp = fopen(optfile_name,"r");
+    } else {                          /* if '@file.opt' */
+        fp = fopen(fname,"r");
     }
-    fp = fopen(optfile_name,"r");
-    
-    if(fp==NULL) {
-	opt_warn_1("Cannot open options file [%s]\n",optfile_name);
-	return 0;
+    if (fp == NULL) {
+        return 0;
     }
-    
+
+    tmp_fileflag = opt_fileflag;
+    opt_fileflag = True;
     while( opt_getline(line,fp) > 0 )
     	opt_lineprocess(line);
+    opt_fileflag = tmp_fileflag;
     
     fclose(fp);
     return 1;                   /* success */
