@@ -1,4 +1,24 @@
-/* $Id: opt_reg.c,v 3.6 1997/12/13 05:53:34 jt Exp $ */
+/* OPT v3.7: options parsing tool */
+/*
+ * Copyright (C) 1998, James Theiler (jt@lanl.gov)
+ * 
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License as published by the Free Software Foundation; either
+ * version 2 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the
+ * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+ * Boston, MA  02111-1307, USA.
+ * 
+ */
+/* $Id: opt_reg.c,v 3.12 1998/07/26 22:25:03 jt Exp $ */
 /* opt_reg.c */
 
 /************************ register options ********************************
@@ -47,6 +67,17 @@
 #include "opt_p.h"
 #endif
 
+/* Default is to indicate whether a flag is true or false by using + or -
+ * But, by #define'ing FLAG_ZERO_ONE, we will use 1 or 0 instead
+ */
+#ifdef FLAG_ZERO_ONE
+#define FLAG_TRUE '1'
+#define FLAG_FALSE '0'        
+#else
+#define FLAG_TRUE '+'
+#define FLAG_FALSE '-'        
+#endif
+
 /* Prototypes for static (local) functions */
 static void opt_fileheader(FILE *);
 static int opt_action(int, ARGVECTOR *);
@@ -91,7 +122,9 @@ int              opt_nreg=0;                     /* number of registered opts */
 #define	OPT_isflagtype(o) \
 	((o)==OPT_FLAG ||    (o)==OPT_NEGFLAG || \
          (o)==OPT_ABSFLAG || (o)==OPT_ABSNEGFLAG)
-#define	OPT_isundelimtype(o) ((o)==OPT_UNDELIMV || (o)==OPT_UNDELIMC)
+#define	OPT_isundelimtype(o) ((o)==OPT_UNDELIM ||  \
+                              (o)==OPT_UNDELIMC || \
+                              (o)==OPT_UNDELIMV )
 
 	/* -------------------------------------------------- */
 	/*    OPTVALUE: a macro to get the value of an option */
@@ -118,7 +151,7 @@ int optregcb(void *val, opt_TYPE otype, char name, char *descript) {
     return optrega(val,otype,name,NULL,descript);
 }
 int optregs(void *val, opt_TYPE otype, char *longname) {
-    return optrega(val,otype,'\0',longname,NULL);
+    return optrega(val,otype,'\0',longname,longname);
 }
 int optregsb(void *val, opt_TYPE otype, char *longname, char *descript) {
     return optrega(val,otype,'\0',longname,descript);
@@ -167,7 +200,8 @@ optrega(void *val, opt_TYPE otype, char name, char *longname, char *descript)
      * undelimted option by setting the name to '\0'
      */
     if(name=='\0' && ISEMPTYSTRING(longname)) {
-        if(otype==OPT_VSTRING) otype=OPT_UNDELIMV;
+        if(otype==OPT_STRING)  otype=OPT_UNDELIM;
+        if(otype==OPT_VSTRING) otype=OPT_UNDELIM;
         if(otype==OPT_CSTRING) otype=OPT_UNDELIMC;
     }
 
@@ -187,7 +221,8 @@ optrega(void *val, opt_TYPE otype, char name, char *longname, char *descript)
     
     if (val != NULL) {
         if(otype==OPT_UNDELIMC) *((char *)optlist[opt_nreg].value)='\0';
-        if(otype==OPT_UNDELIMV) SETOPTVALUE(char *,opt_nreg,NULL);
+        if(otype==OPT_UNDELIM ||
+           otype==OPT_UNDELIMV) SETOPTVALUE(char *,opt_nreg,NULL);
     }
     ++opt_nreg;
     return opt_nreg-1;
@@ -330,24 +365,34 @@ optstrval(int i)
     case OPT_INT:
         sprintf(stval_buf,"%d", OPTVALUE(int,i) );
         break;
+    case OPT_UINT:
     case OPT_UNSINT:
         sprintf(stval_buf,"%u", OPTVALUE(unsigned int,i) );
         break;
     case OPT_SHORT:
         sprintf(stval_buf,"%d", OPTVALUE(short,i) );
         break;
+    case OPT_USHORT:
+        sprintf(stval_buf,"%u", OPTVALUE(unsigned short,i) );
+        break;
     case OPT_LONG:
         sprintf(stval_buf,"%ld", OPTVALUE(long,i) );
         break;
+    case OPT_ULONG:
+        sprintf(stval_buf,"%lu", OPTVALUE(unsigned long,i) );
+        break;
+    case OPT_CHAR:
+        sprintf(stval_buf,"%c", OPTVALUE(char,i));
+        break;
+    case OPT_UCHAR:
+        sprintf(stval_buf,"%c", OPTVALUE(unsigned char,i));
+        break;
     case OPT_FLOAT:
         sprintf(stval_buf,"%g", OPTVALUE(float,i) );
         break;
     case OPT_DOUBLE:
         sprintf(stval_buf,"%g", OPTVALUE(double,i) );
         break;
-    case OPT_CHAR:
-        sprintf(stval_buf,"%c", OPTVALUE(char,i));
-        break;
     case OPT_INTLEVEL:
         if (1) {
             int j;
@@ -362,19 +407,21 @@ optstrval(int i)
     case OPT_FLAG:
     case OPT_ABSFLAG:
         maybe = OPTVALUE(int,i);
-        sprintf(stval_buf,"%c", (maybe==OPT_TRUE ? '+' : '-') );
+        sprintf(stval_buf,"%c", (maybe==OPT_TRUE ? FLAG_TRUE : FLAG_FALSE) );
         break;
     case OPT_NEGFLAG:
     case OPT_ABSNEGFLAG:
         maybe = OPTVALUE(int,i);
-        sprintf(stval_buf,"%c", (maybe==OPT_TRUE ? '-' : '+') );
+        sprintf(stval_buf,"%c", (maybe==OPT_TRUE ? FLAG_FALSE : FLAG_TRUE) );
         break;
     case OPT_CSTRING:    
     case OPT_UNDELIMC:
         /* OPTVALUE(char *,i) macro doesn't work for CSTRING's */
         sprintf(stval_buf,"\"%s\"",(char *)optlist[i].value);
         break;
+    case OPT_STRING:
     case OPT_VSTRING:
+    case OPT_UNDELIM:
     case OPT_UNDELIMV:
         if( optlist[i].value == NULL )	/* this should never happen */
             sprintf(stval_buf,"\"\"");
@@ -407,35 +454,37 @@ optstrtyp(int i)
     static char	sttyp_buf[80];
 
     switch( optlist[i].type ) {
-    case OPT_INT:
-    case OPT_UNSINT:
-    case OPT_SHORT:
-    case OPT_LONG:
-        strcpy(sttyp_buf,"<integer>");
-        break;
-    case OPT_FLOAT:
-    case OPT_DOUBLE:
-        strcpy(sttyp_buf,"<real>");
-        break;
-    case OPT_CHAR:
-        strcpy(sttyp_buf,"<character>");
-        break;
-    case OPT_INTLEVEL:
-    case OPT_FLAG:
-    case OPT_NEGFLAG:
-    case OPT_ABSFLAG:
-    case OPT_ABSNEGFLAG:
-        strcpy(sttyp_buf,"<flag>");
-        break;
-    case OPT_VSTRING:
-    case OPT_CSTRING:
-    case OPT_UNDELIMV:
-    case OPT_UNDELIMC:
-        strcpy(sttyp_buf,"<string>");
-        break;
-    case OPT_NUL:
-        strcpy(sttyp_buf,"<nul>");
-        break;
+        /* Integers */
+    case OPT_INT:         strcpy(sttyp_buf,"<INT>"); break;
+    case OPT_UNSINT:      /* deprecated, same as UINT */
+    case OPT_UINT:        strcpy(sttyp_buf,"<UINT>"); break;
+    case OPT_SHORT:       strcpy(sttyp_buf,"<SHORT>"); break;
+    case OPT_LONG:        strcpy(sttyp_buf,"<LONG>"); break;
+    case OPT_USHORT:      strcpy(sttyp_buf,"<USHORT>"); break;
+    case OPT_ULONG:       strcpy(sttyp_buf,"<ULONG>"); break;
+        /* Floating point */
+    case OPT_FLOAT:       strcpy(sttyp_buf,"<FLOAT>"); break;
+    case OPT_DOUBLE:      strcpy(sttyp_buf,"<DOUBLE>"); break;
+        /* Char */
+    case OPT_CHAR:        strcpy(sttyp_buf,"<CHAR>"); break;
+    case OPT_UCHAR:       strcpy(sttyp_buf,"<UCHAR>"); break;
+        /* Flag */
+    case OPT_INTLEVEL:    strcpy(sttyp_buf,"<INTLEVEL>"); break;
+    case OPT_FLAG:        strcpy(sttyp_buf,"<FLAG>"); break;
+    case OPT_NEGFLAG:     strcpy(sttyp_buf,"<NEGFLAG>"); break;
+    case OPT_ABSFLAG:     strcpy(sttyp_buf,"<ABSFLAG>"); break;
+    case OPT_ABSNEGFLAG:  strcpy(sttyp_buf,"<ABSNEGFLAG>"); break;
+        /* String */
+    case OPT_STRING:      strcpy(sttyp_buf,"<STRING>"); break;
+    case OPT_VSTRING:     strcpy(sttyp_buf,"<STRING>"); break;
+    case OPT_CSTRING:     strcpy(sttyp_buf,"<STRING>"); break;
+        /* Undelimited */
+    case OPT_UNDELIM:     strcpy(sttyp_buf,"<UNDELIM>"); break;
+    case OPT_UNDELIMV:    strcpy(sttyp_buf,"<UNDELIM>"); break;
+    case OPT_UNDELIMC:    strcpy(sttyp_buf,"<UNDELIM>"); break;
+        /* Nul */
+    case OPT_NUL:         strcpy(sttyp_buf,"<NUL>"); break;
+        
     default:
         opt_fatal("usage: undefined o-type");
     }
@@ -472,11 +521,33 @@ opt_tofile(FILE *fp)
             format = fnumstr;
         fprintf(fp,format,optlist[i].name,
                 optstrval(i),IGNOREEOL,optlist[i].descript);
+        /* If an option has both a short and a long option name
+         * then write it as a comment '    ; -b == --option-b
+         */
         if (optlist[i].longname) {
             fprintf(fp,"%-42s%c -%c == --%s\n","",IGNOREEOL,
                     optlist[i].name,optlist[i].longname);
         }
     }
+    /* Now Long Option Names (thanks to Jason V. Morgan) */
+    for(i=0; i<opt_nreg; ++i) {
+        /* do not output the long name if a short name exists */
+        if (optlist[i].name != '\0')
+           continue;
+        if (!optlist[i].longname) 
+           continue;
+
+        /* The following arguably applies to short options too ?? */
+        /* don't want anything with a hook (do we?) */
+        if (optlist[i].type == OPT_NUL)
+            continue;
+        if (optlist[i].hook != NULL)
+            continue;
+        
+        fprintf(fp, "--%s=%-*s %c%s\n", optlist[i].longname, 
+                38 - strlen(optlist[i].longname), optstrval(i), 
+                IGNOREEOL, optlist[i].descript);
+    }
 }
 static void
 opt_fileheader(FILE *fp)
@@ -508,7 +579,7 @@ opt_mstring(int i)
     char	*s;
     char	*optstrval(int i);
     char	*format = "%c %-40s %s";
-    char        *lformat = "  --%s";
+    char        *lformat = "  --%s %*s%s";
 
     if( OPT_isundelimtype( optlist[i].type ) )
 		return NULL;
@@ -534,9 +605,9 @@ opt_mstring(int i)
         }
         s = flgstr;
     } else if( OPT_isflagtype( optlist[i].type ) ) {
-        if( *s == '+' )
+        if( *s == '+' || *s == '1')
             strcpy(flgstr,"TRUE");
-        if( *s == '-' )
+        if( *s == '-' || *s == '0')
             strcpy(flgstr,"FALSE");
         s=flgstr;
     }
@@ -547,7 +618,8 @@ opt_mstring(int i)
     if (optlist[i].name) {
         sprintf(mstring,format,optlist[i].name,optlist[i].descript,s);
     } else if (optlist[i].longname) {
-        sprintf(mstring,lformat,optlist[i].longname);
+        sprintf(mstring,lformat,optlist[i].longname,
+                38 - strlen(optlist[i].longname),"",s);
     }
             
     return mstring;
@@ -599,6 +671,14 @@ opt_usage(void)
             }
         }
     }
+    if (opt_additional_usage_fcn != NULL) {
+        /* precede additional usage by -- to indicate that we
+         * are finished with the options
+         */
+        opt_message(" --\n");
+        /* return value is ignored */
+        (*opt_additional_usage_fcn)();
+    }
 }
 
 	/* ----------------------- */
@@ -633,7 +713,8 @@ opt_undelim(ARGVECTOR *ag)
             optlist[i].invoked += 1;
             return 1; /* success */
         }
-        if (optlist[i].type==OPT_UNDELIMV ) {
+        if (optlist[i].type==OPT_UNDELIM ||
+            optlist[i].type==OPT_UNDELIMV) {
             /** optlist[i].value = strdup(s); <-- this fails! **/
             SETOPTVALUE(char *,i,strdup(s));
             optlist[i].invoked += 1;
@@ -711,9 +792,10 @@ opt_action(int i, ARGVECTOR *ag)
 
     switch( o=optlist[i].type ) {
     case OPT_INT:
-        SETOPTVALUE(int,i, argnextnum(ag));
+        SETOPTVALUE(int,i, argnextnum(ag)); 
         break;
-    case OPT_UNSINT:
+    case OPT_UNSINT: /* deprecated */
+    case OPT_UINT:
         SETOPTVALUE(unsigned int,i, argnextnum(ag));
         break;
     case OPT_SHORT:
@@ -722,6 +804,12 @@ opt_action(int i, ARGVECTOR *ag)
     case OPT_LONG:
         SETOPTVALUE(long,i, argnextnum(ag));
         break;
+    case OPT_USHORT:
+        SETOPTVALUE(unsigned short,i, argnextnum(ag));
+        break;
+    case OPT_ULONG:
+        SETOPTVALUE(unsigned long,i, argnextnum(ag));
+        break;
     case OPT_FLOAT:
         SETOPTVALUE(float,i, argnextnum(ag));
         break;
@@ -731,6 +819,9 @@ opt_action(int i, ARGVECTOR *ag)
     case OPT_CHAR:
         SETOPTVALUE(char,i, ag_c_advance(ag));
         break;
+    case OPT_UCHAR:
+        SETOPTVALUE(unsigned char,i, ag_c_advance(ag));
+        break;
 
     case OPT_FLAG:
     case OPT_NEGFLAG:
@@ -752,10 +843,12 @@ opt_action(int i, ARGVECTOR *ag)
         c = ag_c(ag);           /* character following 'c' */
         switch(c) {
         case '+':
+        case '1':
             SETOPTVALUE(int,i,yes);
             ag_c_advance(ag);       /* eat the '+' */
             break;
         case '-':
+        case '0':
             SETOPTVALUE(int,i,no);
             ag_c_advance(ag);       /* eat the '-' */
             break;
@@ -790,6 +883,7 @@ opt_action(int i, ARGVECTOR *ag)
         }
         break;
         
+    case OPT_STRING:
     case OPT_VSTRING:
         s = argnext(ag);
         SETOPTVALUE(char *,i,strdup(s));
